from analogai.foundation.common.utils.time_utils import Tense


class VerbConjugator:
    @staticmethod
    def conjugate(verb: str, tense: Tense, is_plural: bool = False) -> str:
        if tense == Tense.PAST:
            return VerbConjugator._past_tense(verb, is_plural)
        elif tense == Tense.FUTURE:
            return VerbConjugator._future_tense(verb, is_plural)
        else:
            return verb
    
    @staticmethod
    def _past_tense(verb: str, is_plural: bool) -> str:
        past_forms = {
            "is": "was",
            "are": "were",
            "has": "had",
            "have": "had",
            "does": "did",
            "do": "did",
            "causes": "caused",
            "cause": "caused",
            "is not": "was not",
            "are not": "were not",
            "does not have": "did not have",
            "do not have": "did not have",
            "does not": "did not",
            "do not": "did not",
            "is related to": "was related to",
            "are related to": "were related to",
            "is capable of": "was capable of",
            "are capable of": "were capable of",
            "is part of": "was part of",
            "are part of": "were part of",
            "might be": "might have been",
            "might have": "might have had",
        }
        
        if verb in past_forms:
            return past_forms[verb]
        
        if verb.startswith("is "):
            return verb.replace("is ", "was ", 1)
        if verb.startswith("are "):
            return verb.replace("are ", "were ", 1)
        if verb.startswith("has "):
            return verb.replace("has ", "had ", 1)
        if verb.startswith("have "):
            return verb.replace("have ", "had ", 1)
        if verb.startswith("does "):
            return verb.replace("does ", "did ", 1)
        if verb.startswith("do "):
            return verb.replace("do ", "did ", 1)
        
        if verb.endswith("s") and not is_plural and " " not in verb:
            return verb[:-1] + "ed" if len(verb) > 1 else verb + "ed"
        
        if verb.endswith("e") and " " not in verb:
            return verb + "d"
        
        if " " not in verb:
            return verb + "ed"
        
        return verb
    
    @staticmethod
    def _future_tense(verb: str, is_plural: bool) -> str:
        future_forms = {
            "is": "will be",
            "are": "will be",
            "has": "will have",
            "have": "will have",
            "does": "will",
            "do": "will",
            "causes": "will cause",
            "cause": "will cause",
            "is not": "will not be",
            "are not": "will not be",
            "does not have": "will not have",
            "do not have": "will not have",
            "does not": "will not",
            "do not": "will not",
            "is related to": "will be related to",
            "are related to": "will be related to",
            "is capable of": "will be capable of",
            "are capable of": "will be capable of",
            "is part of": "will be part of",
            "are part of": "will be part of",
        }
        
        if verb in future_forms:
            return future_forms[verb]
        
        if verb.startswith("is "):
            return verb.replace("is ", "will be ", 1)
        if verb.startswith("are "):
            return verb.replace("are ", "will be ", 1)
        if verb.startswith("has "):
            return verb.replace("has ", "will have ", 1)
        if verb.startswith("have "):
            return verb.replace("have ", "will have ", 1)
        if verb.startswith("does "):
            return verb.replace("does ", "will ", 1)
        if verb.startswith("do "):
            return verb.replace("do ", "will ", 1)
        if verb.startswith("might "):
            return verb.replace("might ", "will ", 1)
        
        if " " not in verb:
            return f"will {verb}"
        
        return f"will {verb}"
    
    @staticmethod
    def to_infinitive(verb: str) -> str:
        infinitive_forms = {
            "is": "be",
            "are": "be",
            "was": "be",
            "were": "be",
            "has": "have",
            "have": "have",
            "had": "have",
            "does": "do",
            "do": "do",
            "did": "do",
            "causes": "cause",
            "cause": "cause",
            "caused": "cause",
            "is not": "not be",
            "are not": "not be",
            "was not": "not be",
            "were not": "not be",
            "does not have": "not have",
            "do not have": "not have",
            "did not have": "not have",
            "does not": "not do",
            "do not": "not do",
            "did not": "not do",
            "is related to": "be related to",
            "are related to": "be related to",
            "was related to": "be related to",
            "were related to": "be related to",
            "is capable of": "be capable of",
            "are capable of": "be capable of",
            "was capable of": "be capable of",
            "were capable of": "be capable of",
            "is part of": "be part of",
            "are part of": "be part of",
            "was part of": "be part of",
            "were part of": "be part of",
            "might be": "be",
            "might have": "have",
            "will be": "be",
            "will have": "have",
            "will": "do",
            "will not be": "not be",
            "will not have": "not have",
            "will not": "not do",
        }
        
        if verb in infinitive_forms:
            return infinitive_forms[verb]
        
        if verb.startswith("is "):
            return verb.replace("is ", "be ", 1)
        if verb.startswith("are "):
            return verb.replace("are ", "be ", 1)
        if verb.startswith("was "):
            return verb.replace("was ", "be ", 1)
        if verb.startswith("were "):
            return verb.replace("were ", "be ", 1)
        if verb.startswith("has "):
            return verb.replace("has ", "have ", 1)
        if verb.startswith("have "):
            return verb.replace("have ", "have ", 1)
        if verb.startswith("had "):
            return verb.replace("had ", "have ", 1)
        if verb.startswith("does "):
            return verb.replace("does ", "do ", 1)
        if verb.startswith("do "):
            return verb.replace("do ", "do ", 1)
        if verb.startswith("did "):
            return verb.replace("did ", "do ", 1)
        if verb.startswith("will "):
            return verb.replace("will ", "", 1)
        if verb.startswith("will not "):
            return "not " + verb.replace("will not ", "", 1)
        
        if verb.endswith("s") and " " not in verb:
            return verb[:-1]
        
        if verb.endswith("ed") and " " not in verb:
            if len(verb) > 2 and verb[-3] in "aeiou":
                return verb[:-1]
            return verb[:-2]
        
        return verb


__all__ = ["VerbConjugator"]



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Callable, List, Optional, Protocol


class FactResolver(Protocol):
    def resolve(self, fact_id: str) -> Optional["BaseFact"]:
        raise NotImplementedError


@dataclass
class BaseFact:
    id: str
    source_id: str
    turn_id: Optional[str] = None
    user_id: Optional[str] = None
    agent_id: Optional[str] = None
    embedding: Optional[List[float]] = None
    prev_id: Optional[str] = None
    next_id: Optional[str] = None
    created_at: Optional[datetime] = None
    _resolver: Optional[FactResolver] = field(default=None, repr=False, compare=False)
    _deleter: Optional[Callable[[str], None]] = field(default=None, repr=False, compare=False)
    _statement_service: Optional[object] = field(default=None, repr=False, compare=False)
    _notion_service: Optional[object] = field(default=None, repr=False, compare=False)

    def __post_init__(self) -> None:
        if not self.id:
            raise ValueError("BaseFact requires id")
        if not self.source_id or not str(self.source_id).strip():
            raise ValueError("BaseFact requires source_id")

    def prev(self) -> Optional["BaseFact"]:
        if self._resolver is None or not self.prev_id:
            return None
        return self._resolver.resolve(self.prev_id)

    def next(self) -> Optional["BaseFact"]:
        if self._resolver is None or not self.next_id:
            return None
        return self._resolver.resolve(self.next_id)

    def delete(self) -> None:
        self._delete_fact()

    def _delete_fact(self) -> None:
        if not self.id:
            raise ValueError("BaseFact requires id")
        if self._deleter is None:
            raise NotImplementedError("BaseFact delete requires deleter")
        self._deleter(self.id)


__all__ = ["BaseFact", "FactResolver"]



from dataclasses import dataclass
from typing import Union, Optional

from analogai.foundation.common.enums.relationship_type import RelationshipType
from analogai.foundation.common.utils.relation_utils import normalize_relation_value


@dataclass(frozen=True)
class Relation:
    value: str
    _relationship_type: Optional[RelationshipType]

    @classmethod
    def from_type(cls, relationship_type: RelationshipType) -> 'Relation':
        return cls(
            value=relationship_type.value,
            _relationship_type=relationship_type
        )

    @classmethod
    def from_string(cls, value: str) -> 'Relation':
        if not value or value.strip() == "":
            raise ValueError("relation cannot be empty")
        cleaned_value = value.strip()
        for rel_type in RelationshipType:
            if rel_type.value == cleaned_value:
                return cls(value=cleaned_value, _relationship_type=rel_type)
        normalized_value = normalize_relation_value(cleaned_value)
        if normalized_value != cleaned_value:
            for rel_type in RelationshipType:
                if rel_type.value == normalized_value:
                    return cls(value=cleaned_value, _relationship_type=rel_type)
        return cls(value=cleaned_value, _relationship_type=None)

    @classmethod
    def create(cls, value: Union[str, RelationshipType]) -> 'Relation':
        if isinstance(value, RelationshipType):
            return cls.from_type(value)
        return cls.from_string(value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Relation):
            return self.value == other.value
        raise NotImplementedError(f"Cannot compare Relation with {type(other)}")

    def is_of_type(self, relationship_type: RelationshipType) -> bool:
        if self._relationship_type is not None:
            return self._relationship_type == relationship_type
        return self.value == relationship_type.value

    def __hash__(self):
        return hash(self.value)

    def __str__(self):
        return self.value



from unittest import TestCase

from analogai.foundation.settings import config
from analogai.foundation.modules.fact.factories.fact_factory import FactServiceFactory
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig
from analogai.foundation.modules.fact.source.fact_source import UrlSource
from analogai.foundation.modules.fact.repositories.infrastructure.in_memory_fact_repository import (
    InMemoryFactRepository,
)
from analogai.foundation.modules.fact.source.services.fact_source_service import FactSourceService
from analogai.foundation.modules.fact.source.repositories.infrastructure.in_memory_fact_source_repository import (
    InMemoryFactSourceRepository,
)


class TestFactServiceFunctional(TestCase):
    def setUp(self):
        self._config_override = config.override(
            **{"vector_search.enabled": False, "ENVIRONMENT": "development"}
        )
        self._config_override.__enter__()
        self.config = ServicesTestConfig()
        self.user_id = self.config.user.id
        self.agent_id = self.config.agent.id
        self.subject_notion = self.config.notion_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            expression=NotionExpression(caption="human"),
        )
        self.complement_notion = self.config.notion_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            expression=NotionExpression(caption="mortal"),
        )
        self.source_repo = InMemoryFactSourceRepository()
        self.fact_repo = InMemoryFactRepository()
        self.fact_factory = FactServiceFactory(
            repository=self.fact_repo,
            source_repository=self.source_repo,
            statement_service=self.config.statement_service,
            notion_service=self.config.notion_service,
        )
        self.fact_service = self.fact_factory.create()
        self.fact_registry = self.fact_factory.get_observer_registry()

    def tearDown(self):
        if self.config:
            self.config.tear_down()
        if self._config_override:
            self._config_override.__exit__(None, None, None)

    def test_record_fact_registers_source_and_facts(self):
        source = UrlSource("https://example.com", agent_id=self.agent_id)
        statement = self.config.statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.9,
            validity=0.9,
        )
        fact_id = self.fact_service.record_direct_statement_fact(
            user_id=self.user_id,
            agent_id=self.agent_id,
            statement_id=statement.id,
            text="Humans are mortal. They think.",
            role="user",
            source=source,
        )
        self.assertTrue(fact_id, "Expected fact to be recorded")
        facts = self.fact_service.get_facts_by_source(source.id)
        self.assertTrue(facts, "Expected facts to be recorded")
        sources = self.source_repo.list_sources_for_agent(self.agent_id)
        self.assertIn(source.id, {item.id for item in sources})
        self.assertIsNotNone(self.source_repo.get_source(source.id))

    def test_record_fact_registers_source_via_service(self):
        source = UrlSource("https://example.com/service", agent_id=self.agent_id)
        source_service = FactSourceService(
            fact_service=self.fact_service,
            belief_service=None,
            source_repository=self.source_repo,
        )
        self.fact_service.set_source_service(source_service)

        statement = self.config.statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.9,
            validity=0.9,
        )
        fact_id = self.fact_service.record_direct_statement_fact(
            user_id=self.user_id,
            agent_id=self.agent_id,
            statement_id=statement.id,
            text="Service-backed source registration.",
            role="user",
            source=source,
        )

        self.assertTrue(fact_id, "Expected fact to be recorded")
        self.assertIsNotNone(self.source_repo.get_source(source.id))

    def test_observer_records_fact_with_source_on_statement_create(self):
        source = UrlSource("https://example.com/observer", agent_id=self.agent_id)
        sentence = "Humans are mortal."
        self.fact_registry.append_current_sentence(
            text=sentence,
            role="user",
            source=source,
        )
        statement = self.config.statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            subject_notion=self.subject_notion,
            complement_notion=self.complement_notion,
            relation=Relation.from_string("is"),
            probability=0.9,
            validity=0.9,
        )

        facts = self.fact_service.get_facts_by_source(source.id)
        self.assertTrue(facts, "Expected observer to record a fact")
        recorded = facts[0]
        self.assertEqual(recorded.statement_id, statement.id)
        self.assertEqual(getattr(recorded, "text", ""), sentence)

    def test_observer_records_fact_with_source_on_notion_create(self):
        source = UrlSource("https://example.com/observer-notion", agent_id=self.agent_id)
        sentence = "A human is a mammal."
        self.fact_registry.append_current_sentence(
            text=sentence,
            role="user",
            source=source,
        )
        notion = self.config.notion_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            expression=NotionExpression(caption="mammal"),
        )

        facts = self.fact_service.get_facts_by_source(source.id)
        self.assertTrue(facts, "Expected observer to record a fact")
        recorded = facts[0]
        self.assertEqual(recorded.notion_id, notion.id)
        self.assertEqual(getattr(recorded, "text", ""), sentence)


if __name__ == "__main__":
    import unittest

    unittest.main()



from __future__ import annotations

from datetime import datetime
from typing import Optional, List

from analogai.foundation.modules.fact.basefact import BaseFact
from analogai.foundation.modules.fact.enums.fact_reference_type import FactReferenceType
from analogai.foundation.modules.fact.notion_fact import NotionFact
from analogai.foundation.modules.fact.direct_statement_fact import DirectStatementFact
from analogai.foundation.modules.fact.hypothetical_statement_fact import (
    HypotheticalStatementFact,
)


def build_fact(
    *,
    id: str,
    turn_id: Optional[str] = None,
    statement_id: Optional[str] = None,
    notion_id: Optional[str] = None,
    reference_type: Optional[str] = None,
    user_id: Optional[str] = None,
    agent_id: Optional[str] = None,
    text: str = "",
    embedding: Optional[List[float]] = None,
    source_id: Optional[str] = None,
    prev_id: Optional[str] = None,
    next_id: Optional[str] = None,
    created_at: Optional[datetime] = None,
) -> BaseFact:
    if not source_id or not str(source_id).strip():
        raise ValueError("Fact requires source_id")
    ref = str(reference_type).upper() if reference_type else None
    if ref == FactReferenceType.STATEMENT.value:
        if not statement_id:
            raise ValueError("DirectStatementFact requires statement_id")
        fact = DirectStatementFact(
            id=id,
            turn_id=turn_id,
            statement_id=statement_id,
            user_id=user_id,
            agent_id=agent_id,
            embedding=embedding,
            source_id=source_id,
            prev_id=prev_id,
            next_id=next_id,
            created_at=created_at,
        )
        if text:
            setattr(fact, "text", text)
        return fact
    if ref == FactReferenceType.HYPOTHETICAL_STATEMENT.value:
        if not statement_id:
            raise ValueError("HypotheticalStatementFact requires statement_id")
        fact = HypotheticalStatementFact(
            id=id,
            turn_id=turn_id,
            statement_id=statement_id,
            user_id=user_id,
            agent_id=agent_id,
            embedding=embedding,
            source_id=source_id,
            prev_id=prev_id,
            next_id=next_id,
            created_at=created_at,
        )
        if text:
            setattr(fact, "text", text)
        return fact
    if ref == FactReferenceType.NOTION.value:
        if not notion_id:
            raise ValueError("NotionFact requires notion_id")
        fact = NotionFact(
            id=id,
            turn_id=turn_id,
            notion_id=notion_id,
            user_id=user_id,
            agent_id=agent_id,
            embedding=embedding,
            source_id=source_id,
            prev_id=prev_id,
            next_id=next_id,
            created_at=created_at,
        )
        if text:
            setattr(fact, "text", text)
        return fact
    if statement_id:
        fact = DirectStatementFact(
            id=id,
            turn_id=turn_id,
            statement_id=statement_id,
            user_id=user_id,
            agent_id=agent_id,
            embedding=embedding,
            source_id=source_id,
            prev_id=prev_id,
            next_id=next_id,
            created_at=created_at,
        )
        if text:
            setattr(fact, "text", text)
        return fact
    if notion_id:
        fact = NotionFact(
            id=id,
            turn_id=turn_id,
            notion_id=notion_id,
            user_id=user_id,
            agent_id=agent_id,
            embedding=embedding,
            source_id=source_id,
            prev_id=prev_id,
            next_id=next_id,
            created_at=created_at,
        )
        if text:
            setattr(fact, "text", text)
        return fact
    fact = BaseFact(
        id=id,
        turn_id=turn_id,
        user_id=user_id,
        agent_id=agent_id,
        embedding=embedding,
        source_id=source_id,
        prev_id=prev_id,
        next_id=next_id,
        created_at=created_at,
    )
    if text:
        setattr(fact, "text", text)
    return fact


__all__ = ["build_fact"]



from __future__ import annotations

from dataclasses import dataclass

from analogai.foundation.modules.fact.basefact import BaseFact


@dataclass
class HypotheticalStatementFact(BaseFact):
    statement_id: str = ""

    def __post_init__(self) -> None:
        super().__post_init__()
        if not self.statement_id:
            raise ValueError("HypotheticalStatementFact requires statement_id")

    def delete(self) -> None:
        self._delete_fact()


__all__ = ["HypotheticalStatementFact"]



from typing import List, Dict, Any, Optional
from analogai.foundation.infrastructure.storage.query_storage_adapter import QueryStorageAdapter, QueryBuilder
from analogai.foundation.infrastructure.storage.mongodb.mongodb_storage_adapter_impl import MongoDBStorageAdapterImpl
from analogai.foundation.common.logging.app_logger import app_logger
from typeguard import typechecked

class MongoDBQueryStorageAdapterImpl(QueryStorageAdapter):
    
    def __init__(self):
        self.mongodb_adapter = MongoDBStorageAdapterImpl()

    @typechecked
    def store(self, collection: str, data: Any) -> None:
        if hasattr(data, '__dict__'):
            document = data.__dict__.copy()
        else:
            document = dict(data)
        self.mongodb_adapter.store(collection, document)

    @typechecked
    def retrieve_by_id(self, collection: str, id: str) -> Optional[Any]:
        return self.mongodb_adapter.retrieve_by_id(collection, id)

    @typechecked
    def retrieve_by_attributes(self, collection: str, query: Dict[str, Any]) -> List[Any]:
        return self.mongodb_adapter.retrieve_by_attributes(collection, query)

    @typechecked
    def delete(self, collection: str, id: str) -> bool:
        try:
            self.mongodb_adapter.delete(collection, id)
            return True
        except ValueError:
            return False

    @typechecked
    def clear(self, collection: str) -> None:
        self.mongodb_adapter.db[collection].delete_many({})

    @typechecked
    def query(self, collection: str, query_dict: Dict[str, Any]) -> QueryBuilder:
        app_logger.debug(f"Starting MongoDB query builder for {collection}: {query_dict}")
        return QueryBuilder(self, collection, query_dict)

    @typechecked
    def count(self, collection: str, query_dict: Optional[Dict[str, Any]] = None) -> int:
        if query_dict is None:
            query_dict = {}
        return self.mongodb_adapter.db[collection].count_documents(query_dict)



from typing import Optional, Dict, Iterable, List, Tuple, TYPE_CHECKING
from datetime import datetime

from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.common.utils.statement_serializers import serialize_statement
from analogai.foundation.common.utils.relationship_type_storage import select_relation
from analogai.foundation.common.utils.modifier_utils.modifier_matcher import ModifierMatcher
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.categorical_deduction.categorical_deduction import CategoricalDeduction
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.shared.statement_base import StatementBase
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.core.value_objects.attribute import Attribute
if TYPE_CHECKING:
    from analogai.foundation.modules.belief.belief_service import BeliefService


class CategoricalDeductionService:
    def __init__(
        self,
        belief_service: "BeliefService",
    ):
        if belief_service is None:
            raise ValueError("BeliefService is required for categorical deductions")
        self._belief_service = belief_service
        self._matcher = ModifierMatcher()

    def _time_key(self, time: Optional[Time]) -> Optional[Tuple[Optional[int], Optional[int], Optional[int], Optional[int], Optional[int]]]:
        if time is None:
            return None
        return (time.year, time.month, time.day, time.hour, time.minute)

    def _attributes_key(self, attributes: Optional[List[Attribute]]) -> Optional[Tuple[Tuple[Optional[str], Optional[float], Optional[str]], ...]]:
        if attributes is None:
            return None
        return tuple((attribute.tag, attribute.value, attribute.unit) for attribute in attributes)

    def _modifier_has_context(self, modifier: Optional[Modifier]) -> bool:
        if modifier is None:
            return False
        if modifier.location is not None:
            return True
        if modifier.from_time is not None:
            return True
        if modifier.to_time is not None:
            return True
        if modifier.deontic_modality is not None:
            return True
        if modifier.attributes:
            return True
        if modifier.quantity is not None:
            return True
        return False

    def _modifier_key(self, modifier: Optional[Modifier]) -> Optional[Tuple]:
        if not self._modifier_has_context(modifier):
            return None
        assert modifier is not None
        deontic = modifier.deontic_modality.value if modifier.deontic_modality is not None else None
        return (
            modifier.location,
            deontic,
            self._time_key(modifier.from_time),
            self._time_key(modifier.to_time),
            self._attributes_key(modifier.attributes),
            modifier.quantity,
        )

    def _belief_key_from_statement(self, statement: StatementBase) -> Tuple:
        return self._belief_key_from_values(
            agent_id=statement.agent_id,
            subject_id=statement.subject_notion.id,
            object_id=statement.complement_notion.id,
            relation=statement.relation,
            modifier=statement.modifier,
        )

    def _belief_key_from_values(
        self,
        agent_id: str,
        subject_id: str,
        object_id: str,
        relation: Relation,
        modifier: Optional[Modifier],
    ) -> Tuple:
        if not isinstance(relation, Relation):
            raise TypeError("relation must be a Relation instance")
        return (
            agent_id,
            subject_id,
            relation,
            object_id,
            self._modifier_key(modifier),
        )

    def _modifier_exact_matches(self, stored: Optional[Modifier], query: Optional[Modifier]) -> bool:
        return self._matcher.context_exact_matches(stored, query)

    def _belief_matches_query(self, belief: Belief, modifier: Optional[Modifier]) -> bool:
        return self._modifier_exact_matches(belief.modifier, modifier)

    def _index_beliefs_by_subject(self, beliefs: Iterable[Belief]) -> Dict[str, List[Belief]]:
        index: Dict[str, List[Belief]] = {}
        for belief in beliefs:
            index.setdefault(belief.subject_notion.id, []).append(belief)
        return index

    def _build_belief(self, statements: List[StatementBase]) -> Belief:
        if not statements:
            raise ValueError("Cannot build belief without statements")
        ordered = sorted(
            statements,
            key=lambda stmt: getattr(stmt, "created_at", datetime.min),
        )
        first = ordered[0]
        belief = Belief(
            agent_id=first.agent_id,
            id=first.id,
        )
        for statement in ordered:
            belief.add_statement(statement)
        return belief

    def find_deductions_by_agent_id(
        self,
        agent_id: str,
        modifier: Optional[Modifier] = None,
        max_depth: int = 5,
        limit: Optional[int] = None,
        offset: int = 0,
    ) -> List[StatementBase]:
        if self._belief_service is None:
            raise ValueError("BeliefService is not configured")
        getter = getattr(self._belief_service, "get_base_beliefs_for_agent", None)
        if callable(getter):
            try:
                beliefs = getter(agent_id, modifier=modifier, quantity=None)
            except TypeError:
                beliefs = getter(agent_id)
        else:
            beliefs = self._belief_service.get_all_for_agent(agent_id)
        filtered = [
            belief for belief in beliefs
            if self._belief_matches_query(belief, modifier)
        ]
        if offset or limit is not None:
            start = offset
            end = offset + limit if limit is not None else None
            filtered = filtered[start:end]
        beliefs_by_key = {
            self._belief_key_from_values(
                agent_id=belief.agent_id,
                subject_id=belief.subject_notion.id,
                object_id=belief.complement_notion.id,
                relation=belief.relation,
                modifier=belief.modifier,
            ): belief
            for belief in filtered
        }
        return self.resolve_deductions(
            beliefs_by_key=beliefs_by_key,
            modifier=modifier,
            max_depth=max_depth,
        )

    def resolve_deductions(
        self,
        beliefs_by_key: Dict[Tuple, Belief],
        modifier: Optional[Modifier],
        max_depth: int,
    ) -> List[StatementBase]:
        deductions: List[StatementBase] = []
        seen: set[tuple] = set()

        for depth in range(1, max_depth + 1):
            modifier_key = self._modifier_key(modifier)
            # app_logger.info(
            #     "Categorical deduction depth=%s: beliefs=%s, modifier=%s",
            #     depth,
            #     len(beliefs_by_key),
            #     modifier_key,
            # )
            index = self._index_beliefs_by_subject(beliefs_by_key.values())
            new_added = False
            for minor in list(beliefs_by_key.values()):
                if not self._belief_matches_query(minor, modifier):
                    continue
                major_candidates = index.get(minor.complement_notion.id, [])
                for major in major_candidates:
                    if not self._belief_matches_query(major, modifier):
                        continue
                    relation = select_relation(major.relation, minor.relation)
                    if relation is None:
                        continue
                    if minor.subject_notion.id == major.complement_notion.id:
                        continue
                    key = (
                        minor.subject_notion.id,
                        major.complement_notion.id,
                        relation,
                        major.id,
                        minor.id,
                        modifier_key,
                    )
                    if key in seen:
                        continue
                    seen.add(key)
                    # app_logger.info(
                    #     "Categorical deduction depth=%s: minor=%s -> major=%s => relation=%s",
                    #     depth,
                    #     minor.id,
                    #     major.id,
                    #     relation,
                    # )
                    deduction = self.resolve(
                        agent_id=minor.agent_id,
                        major_premise=major,
                        minor_premise=minor,
                        subject_notion=minor.subject_notion,
                        complement_notion=major.complement_notion,
                        relation=relation,
                        modifier=modifier,
                        deduction_id=None,
                    )
                    deductions.append(deduction)
                    belief_key = self._belief_key_from_statement(deduction)
                    if belief_key in beliefs_by_key:
                        beliefs_by_key[belief_key].add_statement(deduction)
                    else:
                        beliefs_by_key[belief_key] = self._build_belief([deduction])
                    new_added = True
            if not new_added:
                # app_logger.info("Categorical deduction stopped at depth=%s: no new deductions", depth)
                break
        return deductions

    def resolve(
        self,
        agent_id: str,
        major_premise: Belief,
        minor_premise: Belief,
        subject_notion: Notion,
        complement_notion: Notion,
        relation: Relation,
        modifier: Optional[Modifier] = None,
        deduction_id: Optional[str] = None,
    ) -> CategoricalDeduction:
        if subject_notion.id == complement_notion.id:
            raise ValueError(
                f"Cannot resolve categorical deduction: source notion '{subject_notion.caption}' (id={subject_notion.id}) "
                "cannot be the same as target notion"
            )

        deduction = CategoricalDeduction(
            agent_id=agent_id,
            major_premise=major_premise,
            minor_premise=minor_premise,
            subject_notion=subject_notion,
            complement_notion=complement_notion,
            relation=relation,
            modifier=modifier,
            id=deduction_id,
        )

        major_serialized = serialize_statement(
            subject_notion_caption=major_premise.subject_notion.caption,
            complement_notion_caption=major_premise.complement_notion.caption,
            relation=major_premise.relation,
            probability=major_premise.probability,
            modifier=major_premise.modifier,
            subject_attributes=major_premise.subject_notion.attributes,
            complement_attributes=major_premise.complement_notion.attributes,
        )
        minor_serialized = serialize_statement(
            subject_notion_caption=minor_premise.subject_notion.caption,
            complement_notion_caption=minor_premise.complement_notion.caption,
            relation=minor_premise.relation,
            probability=minor_premise.probability,
            modifier=minor_premise.modifier,
            subject_attributes=minor_premise.subject_notion.attributes,
            complement_attributes=minor_premise.complement_notion.attributes,
        )
        app_logger.debug(f"Calculated probability: {deduction.probability} and validity: {deduction.validity} from:")
        app_logger.debug(
            f"Major premise: {major_serialized} (prob={major_premise.probability}, valid={major_premise.validity})"
        )
        app_logger.debug(
            f"Minor premise: {minor_serialized} (prob={minor_premise.probability}, valid={minor_premise.validity})"
        )

        deduction_serialized = serialize_statement(
            subject_notion_caption=deduction.subject_notion.caption,
            complement_notion_caption=deduction.complement_notion.caption,
            relation=deduction.relation,
            probability=float(deduction.probability) if deduction.probability is not None else 1.0,
            modifier=deduction.modifier,
            subject_attributes=deduction.subject_notion.attributes,
            complement_attributes=deduction.complement_notion.attributes,
        )
        app_logger.debug(f"Resolved CategoricalDeduction: {deduction_serialized} (id: {deduction.id})")

        return deduction


from __future__ import annotations

from analogai.foundation.infrastructure.embeddings.base import EmbeddingProvider
from analogai.foundation.common.enums.embedding_provider import EmbeddingProvider as EmbeddingProviderEnum
from analogai.foundation.infrastructure.embeddings.local_provider import LocalSentenceTransformerProvider
from analogai.foundation.infrastructure.embeddings.openai_provider import OpenAIEmbeddingProvider
from analogai.foundation.infrastructure.embeddings.azure_openai_provider import AzureOpenAIEmbeddingProvider
from analogai.foundation.infrastructure.embeddings.azure_foundry_provider import AzureFoundryEmbeddingProvider


DEFAULT_LOCAL_MODEL = "all-MiniLM-L6-v2"


def create_embedding_provider(
    provider: EmbeddingProviderEnum,
    model_name: str | None = None,
) -> EmbeddingProvider:
    resolved_model = model_name or DEFAULT_LOCAL_MODEL
    provider_factories = {
        EmbeddingProviderEnum.LOCAL: lambda: LocalSentenceTransformerProvider(
            model_name=resolved_model,
        ),
        EmbeddingProviderEnum.AZURE: lambda: AzureFoundryEmbeddingProvider(
            deployment_name=resolved_model,
        ),
        EmbeddingProviderEnum.AZURE_OPENAI: lambda: AzureOpenAIEmbeddingProvider(
            deployment_name=resolved_model,
        ),
        EmbeddingProviderEnum.AZURE_FOUNDRY: lambda: AzureFoundryEmbeddingProvider(
            deployment_name=resolved_model,
        ),
        EmbeddingProviderEnum.OPENAI: lambda: OpenAIEmbeddingProvider(
            model_name=resolved_model,
        ),
    }

    return provider_factories[provider]()



from typing import Protocol, Optional, Any, TYPE_CHECKING

from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.modifier import Modifier

if TYPE_CHECKING:
    from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement


class BeliefCompositionMember(Protocol):
    id: str
    agent_id: str
    subject_notion: Notion
    complement_notion: Notion
    relation: Any
    modifier: Optional[Modifier]

    @property
    def probability(self) -> float: ...

    @property
    def validity(self) -> float: ...

    @property
    def location(self) -> Optional[str]: ...

    @property
    def from_time(self) -> Any: ...

    @property
    def to_time(self) -> Any: ...

    @property
    def deontic_modality(self) -> Any: ...

    @property
    def quantity(self) -> Optional[float]: ...

    def visit(self, visitor: Any) -> Any: ...



from analogai.foundation.settings import config
from analogai.foundation.extensions.contradictions.contradiction_checker import is_belief_contradictory
from analogai.foundation.extensions.contradictions.probability_variance_calculator import calculate_probability_variance
from analogai.foundation.extensions.contradictions.contradictions_service import ContradictionsService
from analogai.foundation.extensions.contradictions.contradictions_service_factory import ContradictionsServiceFactory

__all__ = [
    "config",
    "is_belief_contradictory",
    "calculate_probability_variance",
    "ContradictionsService",
    "ContradictionsServiceFactory",
]



import uuid
from datetime import datetime
from typing import List, Optional

from analogai.foundation.core.value_objects.cause import Cause
from analogai.foundation.core.value_objects.effect import Effect


class HypotheticalStatement:
    def __init__(
        self,
        user_id: Optional[str],
        agent_id: str,
        causes: List[Cause],
        effect: Effect,
        validity: float = 1.0,
        created_at: Optional[datetime] = None,
        id: Optional[str] = None,
    ):
        self.id = id if id else str(uuid.uuid4())
        self.user_id = user_id
        self.agent_id = agent_id
        self.causes = list(causes or [])
        self.effect = effect
        self.validity = validity
        self.created_at = created_at if created_at else datetime.now()



from analogai.foundation.extensions.contradictions.contradictions_service import ContradictionsService
from analogai.foundation.setup.factories.belief_service_factory import BeliefServiceFactory


class ContradictionsServiceFactory:
    def __init__(self, belief_service=None, environment: str | None = None):
        self.belief_service = belief_service or BeliefServiceFactory(environment=environment).get_service()
        self.service = ContradictionsService(self.belief_service)

    def get_service(self) -> ContradictionsService:
        return self.service



from typing import Optional

from analogai.foundation.common.dtos.categorical_deduction_expression import CategoricalDeductionExpression
from analogai.foundation.common.utils.statement_serializers import serialize_statement
from analogai.foundation.common.utils.statement_serializers.helpers import normalize_timezone


def serialize_categorical_deduction_expression(
    expression: CategoricalDeductionExpression,
    timezone: Optional[str] = None,
) -> str:
    timezone = normalize_timezone(timezone)
    return serialize_statement(
        subject_notion_caption=expression.subject_notion_expression.caption,
        complement_notion_caption=expression.complement_notion_expression.caption,
        relation=expression.relation,
        probability=expression.probability,
        modifier=expression.modifier,
        timezone=timezone,
        subject_attributes=expression.subject_notion_expression.attributes,
        complement_attributes=expression.complement_notion_expression.attributes,
    )



from __future__ import annotations

from typing import Optional

from analogai.foundation.modules.fact.basefact import BaseFact


def format_fact_summary(fact: BaseFact) -> str:
    text = str(getattr(fact, "text", "") or "").strip()
    source = _format_source_id(fact.source_id)
    if text and source:
        return f"source={source}: {text}"
    if text:
        return f"source={fact.user_id}: {text}"
    return f"source={source or fact.user_id}"


def format_fact_with_source(fact: BaseFact) -> str:
    text = get_fact_text(fact).strip()
    if not text:
        return ""
    source = _format_source_id(fact.source_id)
    if not source:
        return text
    return f"{text} (source: {source})"


def get_belief_id(belief) -> Optional[str]:
    if belief is None:
        return None
    value = belief.id
    return str(value) if value else None


def get_statement_id(belief) -> Optional[str]:
    if belief is None:
        return None
    if hasattr(belief, "statement_id"):
        value = belief.statement_id
        return str(value) if value else None
    if hasattr(belief, "statements"):
        statements = belief.statements
        if statements:
            first = statements[0]
            value = getattr(first, "id", None)
            return str(value) if value else None
    return None


def get_fact_text(fact) -> str:
    if fact is None:
        return ""
    return str(getattr(fact, "text", "") or "")


def _format_source_id(source_id: str | None) -> str:
    if not source_id:
        return ""
    value = str(source_id).strip()
    if not value:
        return ""
    return value


def collect_fact_window(
    fact,
    window: int,
    time_window_seconds: int | None = None,
) -> list:
    if fact is None:
        return []
    prevs: list = []
    current = fact
    for _ in range(max(0, int(window))):
        prev_chunk = current.prev()
        if prev_chunk is None:
            break
        prevs.append(prev_chunk)
        current = prev_chunk
    nexts: list = []
    current = fact
    for _ in range(max(0, int(window))):
        next_chunk = current.next()
        if next_chunk is None:
            break
        nexts.append(next_chunk)
        current = next_chunk
    prevs.reverse()
    combined = prevs + [fact] + nexts
    if time_window_seconds is None:
        return combined
    if time_window_seconds <= 0:
        return [fact]
    anchor_time = getattr(fact, "created_at", None)
    if anchor_time is None:
        return combined
    filtered = []
    for item in combined:
        item_time = getattr(item, "created_at", None)
        if item_time is None:
            continue
        delta = abs((item_time - anchor_time).total_seconds())
        if delta <= float(time_window_seconds):
            filtered.append(item)
    return filtered



from .fact_source import FactSource, FileSource, OtherSource, SourceType, UrlSource
from .repositories.fact_source_repository import FactSourceRepository
from .repositories.infrastructure.in_memory_fact_source_repository import InMemoryFactSourceRepository
from .repositories.infrastructure.mongo_fact_source_repository import MongoFactSourceRepository

__all__ = [
    "FactSource",
    "FileSource",
    "OtherSource",
    "SourceType",
    "UrlSource",
    "FactSourceRepository",
    "InMemoryFactSourceRepository",
    "MongoFactSourceRepository",
]



from abc import ABC, abstractmethod
from .observable_service import ObservableService


class HypotheticalStatementObserver(ABC):
    @abstractmethod
    def on_hypothetical_statement_created(self, hypothetical_statement):
        pass


class HypotheticalStatementObservableService(ObservableService):
    def notify_hypothetical_statement_created(self, hypothetical_statement) -> None:
        for observer in self._observers:
            try:
                observer.on_hypothetical_statement_created(hypothetical_statement)
            except (AttributeError, TypeError):
                pass


__all__ = ["HypotheticalStatementObserver", "HypotheticalStatementObservableService"]



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Optional, List, TYPE_CHECKING
from uuid import uuid4

from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.infrastructure.services.embedding_service import EmbeddingService
from analogai.foundation.modules.fact.basefact import BaseFact
from analogai.foundation.modules.fact.fact_deletion_utils import (
    delete_notion,
    delete_statement,
)
from analogai.foundation.modules.fact.notion_fact import NotionFact
from analogai.foundation.modules.fact.direct_statement_fact import DirectStatementFact
from analogai.foundation.modules.fact.hypothetical_statement_fact import (
    HypotheticalStatementFact,
)
from analogai.foundation.modules.fact.source.fact_source import FactSource
from analogai.foundation.modules.fact.repositories.fact_repository import FactRepository
from analogai.foundation.modules.fact.source.repositories.fact_source_repository import FactSourceRepository
from analogai.foundation.modules.fact.enums.fact_search_mode import FactSearchMode
from analogai.foundation.modules.fact.enums.fact_embedding_mode import FactEmbeddingMode

if TYPE_CHECKING:
    from analogai.foundation.modules.fact.source.services.fact_source_service import FactSourceService


class FactServicePort(ABC):
    @abstractmethod
    def record_direct_statement_fact(
        self,
        user_id: str,
        agent_id: str,
        statement_id: str,
        text: str,
        role: str,
        *,
        source: FactSource,
        user_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ) -> str:
        raise NotImplementedError

    @abstractmethod
    def record_notion_fact(
        self,
        user_id: str,
        agent_id: str,
        notion_id: str,
        text: str,
        role: str,
        *,
        source: FactSource,
        user_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ) -> str:
        raise NotImplementedError

    @abstractmethod
    def record_hypothetical_statement_fact(
        self,
        user_id: str,
        agent_id: str,
        statement_id: str,
        text: str,
        role: str,
        *,
        source: FactSource,
        user_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ) -> str:
        raise NotImplementedError

    @abstractmethod
    def link_belief(self, belief_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def link_notion(self, notion_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def link_hypothetical_statement(self, statement_id: str, turn_id: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_turn(self, turn_id: str) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_statement(self, statement_id: str) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_belief(self, belief_id: str) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_notion(self, notion_id: str) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_for_hypothetical_statement(self, statement_id: str) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def get_nearby_turn_facts(
        self,
        turn_id: str,
        user_id: str,
        agent_id: str,
        window: int = 2,
    ) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def search_facts(
        self,
        user_id: str,
        agent_id: str,
        query: str,
        top_k: int = 5,
        similarity_cutoff: Optional[float] = None,
        mode: FactSearchMode = FactSearchMode.VECTOR_THEN_TEXT,
    ) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def get_facts_by_source(self, source_id: str) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def delete_facts_by_source(self, source_id: str) -> List[BaseFact]:
        raise NotImplementedError

    @abstractmethod
    def get_statement_ids_for_turn(self, turn_id: str) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def get_belief_ids_for_turn(self, turn_id: str) -> List[str]:
        raise NotImplementedError


class RepositoryBackedFactService(FactServicePort):
    def __init__(
        self,
        repository: FactRepository,
        source_repository: Optional[FactSourceRepository] = None,
        source_service: Optional["FactSourceService"] = None,
        statement_service: Optional[object] = None,
        notion_service: Optional[object] = None,
        embedding_service: Optional[EmbeddingService] = None,
        embedding_mode: FactEmbeddingMode = FactEmbeddingMode.ENABLED,
        max_chunk_chars: int = 800,
        overlap_sentences: int = 1,
        max_paragraphs_per_fact: Optional[int] = None,
        max_sentences_per_fact: Optional[int] = None,
    ) -> None:
        if statement_service is None or notion_service is None:
            raise ValueError("FactService requires statement_service and notion_service wiring")
        self._repository = repository
        self._source_repository = source_repository
        self._source_service = source_service
        self._statement_service = statement_service
        self._notion_service = notion_service
        self._embedding_service = embedding_service
        self._embedding_mode = embedding_mode
        self._max_chunk_chars = max(200, int(max_chunk_chars))
        self._overlap_sentences = max(0, int(overlap_sentences))
        resolved_paragraphs_per_fact = (
            max_paragraphs_per_fact
            if max_paragraphs_per_fact is not None
            else max_sentences_per_fact
        )
        self._max_paragraphs_per_fact = (
            int(resolved_paragraphs_per_fact)
            if resolved_paragraphs_per_fact is not None
            else None
        )

    def record_direct_statement_fact(
        self,
        user_id: str,
        agent_id: str,
        statement_id: str,
        text: str,
        role: str,
        *,
        source: FactSource,
        user_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ) -> str:
        if not text or not text.strip():
            raise ValueError("Fact text must be provided")
        if not statement_id:
            raise ValueError("Concrete fact requires reference id")
        fact_id = str(uuid4())
        if not isinstance(source, FactSource):
            raise TypeError("Fact source must be a FactSource instance")
        self._register_source(source)
        resolved_timestamp = timestamp or datetime.now(timezone.utc)
        embedding_vector = None
        if self._embedding_mode == FactEmbeddingMode.ENABLED:
            if self._embedding_service is None:
                raise RuntimeError("Fact embedding service unavailable")
            embedding_vector = self._embedding_service.encode(text)
        fact = DirectStatementFact(
            id=fact_id,
            statement_id=statement_id,
            turn_id=None,
            user_id=user_id,
            agent_id=agent_id,
            embedding=embedding_vector,
            source_id=source.id,
            prev_id=None,
            next_id=None,
            created_at=resolved_timestamp,
        )
        setattr(fact, "text", text)
        self._repository.save_fact(fact)
        return fact_id

    def record_notion_fact(
        self,
        user_id: str,
        agent_id: str,
        notion_id: str,
        text: str,
        role: str,
        *,
        source: FactSource,
        user_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ) -> str:
        if not text or not text.strip():
            raise ValueError("Fact text must be provided")
        if not notion_id:
            raise ValueError("Concrete fact requires reference id")
        fact_id = str(uuid4())
        if not isinstance(source, FactSource):
            raise TypeError("Fact source must be a FactSource instance")
        self._register_source(source)
        resolved_timestamp = timestamp or datetime.now(timezone.utc)
        embedding_vector = None
        if self._embedding_mode == FactEmbeddingMode.ENABLED:
            if self._embedding_service is None:
                raise RuntimeError("Fact embedding service unavailable")
            embedding_vector = self._embedding_service.encode(text)
        fact = NotionFact(
            id=fact_id,
            notion_id=notion_id,
            turn_id=None,
            user_id=user_id,
            agent_id=agent_id,
            embedding=embedding_vector,
            source_id=source.id,
            prev_id=None,
            next_id=None,
            created_at=resolved_timestamp,
        )
        setattr(fact, "text", text)
        self._repository.save_fact(fact)
        return fact_id

    def record_hypothetical_statement_fact(
        self,
        user_id: str,
        agent_id: str,
        statement_id: str,
        text: str,
        role: str,
        *,
        source: FactSource,
        user_name: Optional[str] = None,
        timestamp: Optional[datetime] = None,
    ) -> str:
        if not text or not text.strip():
            raise ValueError("Fact text must be provided")
        if not statement_id:
            raise ValueError("Concrete fact requires reference id")
        fact_id = str(uuid4())
        if not isinstance(source, FactSource):
            raise TypeError("Fact source must be a FactSource instance")
        self._register_source(source)
        resolved_timestamp = timestamp or datetime.now(timezone.utc)
        embedding_vector = None
        if self._embedding_mode == FactEmbeddingMode.ENABLED:
            if self._embedding_service is None:
                raise RuntimeError("Fact embedding service unavailable")
            embedding_vector = self._embedding_service.encode(text)
        fact = HypotheticalStatementFact(
            id=fact_id,
            statement_id=statement_id,
            turn_id=None,
            user_id=user_id,
            agent_id=agent_id,
            embedding=embedding_vector,
            source_id=source.id,
            prev_id=None,
            next_id=None,
            created_at=resolved_timestamp,
        )
        setattr(fact, "text", text)
        self._repository.save_fact(fact)
        return fact_id

    def set_source_service(self, source_service: Optional["FactSourceService"]) -> None:
        self._source_service = source_service

    def set_statement_service(self, statement_service: Optional[object]) -> None:
        self._statement_service = statement_service

    def set_notion_service(self, notion_service: Optional[object]) -> None:
        self._notion_service = notion_service

    def _register_source(self, source: FactSource) -> None:
        if source is None:
            return
        if self._source_service is not None:
            try:
                self._source_service.store_source(source)
            except Exception as exc:
                app_logger.warning(
                    "[FactService] Failed to register fact source via service source_id=%s error=%s",
                    source.id,
                    exc,
                )
            return
        if self._source_repository is None:
            return
        try:
            self._source_repository.register_source(source)
        except Exception as exc:
            app_logger.warning(
                "[FactService] Failed to register fact source source_id=%s error=%s",
                source.id,
                exc,
            )

    def link_statement(self, statement_id: str, turn_id: str) -> None:
        self._repository.link_statement(statement_id, turn_id)
        self._repository.update_facts_statement_id(turn_id, statement_id)

    def link_belief(self, belief_id: str, turn_id: str) -> None:
        self._repository.link_belief(belief_id, turn_id)

    def link_notion(self, notion_id: str, turn_id: str) -> None:
        self._repository.link_notion(notion_id, turn_id)
        self._repository.update_facts_notion_id(turn_id, notion_id)

    def link_hypothetical_statement(self, statement_id: str, turn_id: str) -> None:
        self._repository.link_hypothetical_statement(statement_id, turn_id)
        self._repository.update_facts_hypothetical_statement_id(turn_id, statement_id)

    def get_facts_for_turn(self, turn_id: str) -> List[BaseFact]:
        return self._attach_fact_services(self._repository.get_facts_for_turn(turn_id))

    def get_facts_for_statement(self, statement_id: str) -> List[BaseFact]:
        turn_id = self._repository.get_turn_id_for_statement(statement_id)
        return (
            self._attach_fact_services(self._repository.get_facts_for_turn(turn_id))
            if turn_id
            else []
        )

    def get_facts_for_belief(self, belief_id: str) -> List[BaseFact]:
        turn_id = self._repository.get_turn_id_for_belief(belief_id)
        return (
            self._attach_fact_services(self._repository.get_facts_for_turn(turn_id))
            if turn_id
            else []
        )

    def get_facts_for_notion(self, notion_id: str) -> List[BaseFact]:
        turn_id = self._repository.get_turn_id_for_notion(notion_id)
        return (
            self._attach_fact_services(self._repository.get_facts_for_turn(turn_id))
            if turn_id
            else []
        )

    def get_facts_for_hypothetical_statement(self, statement_id: str) -> List[BaseFact]:
        turn_id = self._repository.get_turn_id_for_hypothetical_statement(statement_id)
        return (
            self._attach_fact_services(self._repository.get_facts_for_turn(turn_id))
            if turn_id
            else []
        )

    def get_nearby_turn_facts(
        self,
        turn_id: str,
        user_id: str,
        agent_id: str,
        window: int = 2,
    ) -> List[BaseFact]:
        index = self._repository.get_turn_ids(user_id, agent_id)
        if not index or turn_id not in index:
            return []
        position = index.index(turn_id)
        start = max(0, position - max(0, int(window)))
        end = min(len(index), position + max(0, int(window)) + 1)
        nearby_ids = index[start:end]
        results: List[BaseFact] = []
        for tid in nearby_ids:
            results.extend(self._repository.get_facts_for_turn(tid))
        return self._attach_fact_services(results)

    def search_facts(
        self,
        user_id: str,
        agent_id: str,
        query: str,
        top_k: int = 5,
        similarity_cutoff: Optional[float] = None,
        mode: FactSearchMode = FactSearchMode.VECTOR_THEN_TEXT,
    ) -> List[BaseFact]:
        if not query or not query.strip():
            return []
        if mode == FactSearchMode.TEXT_ONLY:
            return self._attach_fact_services(
                self._repository.search_text(user_id, agent_id, query, top_k=top_k)
            )
        if mode == FactSearchMode.VECTOR_ONLY:
            return self._attach_fact_services(
                self._search_vector_only(user_id, agent_id, query, top_k, similarity_cutoff)
            )
        if mode == FactSearchMode.VECTOR_THEN_TEXT:
            results = self._search_vector_only(user_id, agent_id, query, top_k, similarity_cutoff)
            if results:
                return self._attach_fact_services(results)
            return self._attach_fact_services(
                self._repository.search_text(user_id, agent_id, query, top_k=top_k)
            )
        return []

    def get_facts_by_source(self, source_id: str) -> List[BaseFact]:
        return self._attach_fact_services(self._repository.get_facts_by_source(source_id))

    def delete_facts_by_source(self, source_id: str) -> List[BaseFact]:
        deleted = self._repository.delete_facts_by_source(source_id)
        if deleted:
            deleted = self._attach_fact_services(deleted)
            self._delete_attached_entities(deleted)
        return deleted

    def get_statement_ids_for_turn(self, turn_id: str) -> List[str]:
        return self._repository.get_statement_ids_for_turn(turn_id)

    def get_belief_ids_for_turn(self, turn_id: str) -> List[str]:
        return self._repository.get_belief_ids_for_turn(turn_id)


    def _search_vector_only(
        self,
        user_id: str,
        agent_id: str,
        query: str,
        top_k: int,
        similarity_cutoff: Optional[float],
    ) -> List[BaseFact]:
        if self._embedding_mode != FactEmbeddingMode.ENABLED:
            return []
        if self._embedding_service is None:
            raise RuntimeError("Fact embedding service unavailable")
        query_embedding = self._embedding_service.encode(query)
        if not query_embedding:
            app_logger.warning(
                "[FactService] encode returned empty query embedding query='%s'",
                query.strip(),
            )
            return []
        return self._repository.search_similar(
            user_id,
            agent_id,
            query_embedding,
            top_k=top_k,
            similarity_cutoff=similarity_cutoff,
        )


    def _delete_attached_entities(self, facts: List[BaseFact]) -> None:
        for fact in facts:
            try:
                if isinstance(fact, DirectStatementFact):
                    delete_statement(fact.statement_id, self._statement_service)
                elif isinstance(fact, NotionFact):
                    delete_notion(fact.notion_id, self._notion_service)
            except Exception as exc:
                app_logger.warning(
                    "[FactService] Failed to delete attached entity for fact_id=%s error=%s",
                    getattr(fact, "id", None),
                    exc,
                )

    def _attach_fact_services(self, facts: List[BaseFact]) -> List[BaseFact]:
        if not facts:
            return []
        for fact in facts:
            fact._statement_service = self._statement_service
            fact._notion_service = self._notion_service
            if fact._deleter is None:
                fact._deleter = self._repository.delete_fact
        return facts





import unittest

from analogai.foundation.common.enums.certainty import Certainty
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.core.value_objects.effect import Effect
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.modules.conditionals.hypothetical_statement import HypotheticalStatement


class TestHypotheticalStatementCertainty(unittest.TestCase):
    def _build_effect(self, certainty=None):
        criteria = {
            Criterion.SubjectCaption: "sky",
            Criterion.ObjectCaption: "blue",
            Criterion.Relation: Relation.from_string("is"),
        }
        if certainty is not None:
            criteria[Criterion.Certainty] = certainty
        return Effect(criteria)

    def test_user_id_is_preserved(self):
        hypothetical = HypotheticalStatement(
            user_id="user-1",
            agent_id="agent-1",
            causes=[],
            effect=self._build_effect(),
        )
        self.assertEqual(hypothetical.user_id, "user-1")

    def test_effect_certainty_is_preserved(self):
        hypothetical = HypotheticalStatement(
            user_id="user-1",
            agent_id="agent-1",
            causes=[],
            effect=self._build_effect(Certainty.UNCERTAIN),
        )
        self.assertEqual(
            hypothetical.effect.criteria.get(Criterion.Certainty),
            Certainty.UNCERTAIN,
        )


if __name__ == "__main__":
    unittest.main()



from .event import Event
from analogai.foundation.modules.shared.statement_base import StatementBase


class StatementUpdatedEvent(Event):
    def __init__(self, statement: StatementBase):
        super().__init__(statement)
        self.statement = statement


from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.agent.agent import Agent
from analogai.foundation.modules.agent.agent_service import AgentService
from analogai.foundation.extensions.authorization.core.enums.operation import AuthorizationOperation
from analogai.foundation.common.logging.app_logger import app_logger

class AuthorizationService:
    def __init__(self, agent_service: AgentService):
        self.agent_service = agent_service

    def authorize_user_for_agent(
        self,
        user: User,
        agent: Agent,
        operation: AuthorizationOperation = AuthorizationOperation.CHAT,
    ) -> bool:
        if not user:
            app_logger.error("Authorization failed: No user provided")
            return False
        if not agent:
            app_logger.error("Authorization failed: No agent provided")
            return False
        user_has_access = self._check_user_access_to_agent(user, agent, operation)
        if user_has_access:
            app_logger.info(f"Authorized user {user.id} for agent {agent.id}")
            return True
        app_logger.error(
            f"Authorization failed: User {user.id} does not have access to agent {agent.id}"
        )
        return False

    def find_and_authorize_agent(
        self,
        user: User,
        agent_id: str,
        operation: AuthorizationOperation = AuthorizationOperation.CHAT,
    ) -> Agent:
        if not user:
            app_logger.error("Authorization failed: No user provided")
            raise ValueError("User not authenticated")
        agent = self.agent_service.find(agent_id)
        if not agent:
            app_logger.error(
                f"Authorization failed: Agent with id {agent_id} not found"
            )
            raise ValueError(f"Agent with id {agent_id} not found")
        if not self.authorize_user_for_agent(user, agent, operation):
            raise ValueError(
                f"User {user.id} does not have access to agent {agent_id}"
            )
        return agent

    def _check_user_access_to_agent(
        self,
        user: User,
        agent: Agent,
        operation: AuthorizationOperation,
    ) -> bool:
        if operation == AuthorizationOperation.CHAT:
            return True
        if not agent.roles:
            return False
        for role in agent.roles:
            if role.user and role.user.id == user.id:
                return True
        return False



from __future__ import annotations

from typing import Protocol, Optional, List, Dict, cast

from analogai.foundation.common.utils.statement_serializers import serialize_statement
from analogai.foundation.common.utils.repr_utils import format_statement_repr as format_statement_repr_util
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.core.value_objects.attribute import Attribute
from analogai.foundation.common.enums.relationship_type import RelationshipType


class NotionLike(Protocol):
    id: str
    user_id: str
    agent_id: str
    caption: str
    attributes: Optional[List[Attribute]]


class StatementLike(Protocol):
    subject_notion: NotionLike
    complement_notion: NotionLike
    relation: Relation | RelationshipType | str
    modifier: Optional[Modifier]
    probability: float


class BeliefLike(Protocol):
    id: str
    statements: List[object]
    subject_notion: NotionLike
    complement_notion: NotionLike
    relation: Relation | RelationshipType | str
    modifier: Optional[Modifier]
    probability: float


class AttributeValueFormatter(Protocol):
    def format(self, value: object) -> str:
        ...


class ScalarValueFormatter:
    def format(self, value: object) -> str:
        return str(value)


class ListValueFormatter:
    def format(self, value: object) -> str:
        items = cast(List[object], value)
        return ", ".join(str(item) for item in items)


class AttributeFormatter:
    def __init__(
        self,
        value_formatters: Optional[Dict[type, AttributeValueFormatter]] = None,
        default_formatter: Optional[AttributeValueFormatter] = None,
    ):
        self._value_formatters = value_formatters or {}
        self._default_formatter = default_formatter or ScalarValueFormatter()

    def format(self, attribute: Attribute) -> Optional[str]:
        if attribute.value is None:
            return attribute.tag or None

        formatter = self._value_formatters.get(type(attribute.value), self._default_formatter)
        value_str = formatter.format(attribute.value)
        unit_str = f" {attribute.unit}" if attribute.unit else ""
        if attribute.tag:
            return f"{attribute.tag}={value_str}{unit_str}"
        return f"{value_str}{unit_str}"


class NotionStringFormatter:
    def __init__(self, attribute_formatter: AttributeFormatter):
        self._attribute_formatter = attribute_formatter

    def format(self, notion: NotionLike) -> str:
        if not notion.attributes:
            return notion.caption

        parts: List[str] = []
        for attr in notion.attributes:
            formatted = self._attribute_formatter.format(attr)
            if formatted:
                parts.append(formatted)

        if not parts:
            return notion.caption
        return f"{notion.caption} ({', '.join(parts)})"


_DEFAULT_ATTRIBUTE_FORMATTER = AttributeFormatter(
    value_formatters={list: ListValueFormatter()},
    default_formatter=ScalarValueFormatter(),
)
_DEFAULT_NOTION_FORMATTER = NotionStringFormatter(_DEFAULT_ATTRIBUTE_FORMATTER)


def format_notion_str(notion: NotionLike) -> str:
    return _DEFAULT_NOTION_FORMATTER.format(notion)


def format_notion_repr(notion: NotionLike) -> str:
    return (
        "Notion(id={!r}, user_id={!r}, agent_id={!r}, caption={!r}, attributes={!r})"
    ).format(
        notion.id,
        notion.user_id,
        notion.agent_id,
        notion.caption,
        notion.attributes,
    )


def format_user_repr(user) -> str:
    return (
        "User(" f"id='{user.id}', " f"name='{user.name}', " f"email='{user.email}', " f"timezone='{user.timezone}'" ")"
    )


def format_statement_str(statement: StatementLike) -> str:
    return serialize_statement(
        subject_notion_caption=statement.subject_notion.caption,
        complement_notion_caption=statement.complement_notion.caption,
        relation=statement.relation,
        probability=float(statement.probability),
        modifier=statement.modifier,
        subject_attributes=statement.subject_notion.attributes,
        complement_attributes=statement.complement_notion.attributes,
    )


def format_statement_repr(statement: StatementLike) -> str:
    relation = (
        statement.relation.value
        if isinstance(statement.relation, RelationshipType)
        else statement.relation
    )
    return format_statement_repr_util(
        subject_caption=statement.subject_notion.caption,
        relation=relation,
        object_caption=statement.complement_notion.caption,
        modifier=statement.modifier,
    )


def format_direct_statement_repr(statement: StatementLike) -> str:
    return format_statement_repr(statement)


def format_belief_str(belief: BeliefLike) -> str:
    if not belief.statements:
        return f"Belief(id={belief.id})"
    return serialize_statement(
        subject_notion_caption=belief.subject_notion.caption,
        complement_notion_caption=belief.complement_notion.caption,
        relation=belief.relation,
        probability=float(belief.probability),
        modifier=belief.modifier,
        subject_attributes=belief.subject_notion.attributes,
        complement_attributes=belief.complement_notion.attributes,
    )


def format_belief_repr(belief: BeliefLike) -> str:
    if not belief.statements:
        return f"Belief(id={belief.id})"
    relation = (
        belief.relation.value
        if isinstance(belief.relation, RelationshipType)
        else belief.relation
    )
    return format_statement_repr_util(
        subject_caption=belief.subject_notion.caption,
        relation=relation,
        object_caption=belief.complement_notion.caption,
        modifier=belief.modifier,
    )


def format_categorical_deduction_repr(deduction) -> str:
    return (
        f"CategoricalDeduction(id={deduction.id!r}, agent_id={deduction.agent_id!r}, "
        f"major_premise={deduction.major_premise!r}, minor_premise={deduction.minor_premise!r}, "
        f"subject_notion={deduction.subject_notion!r}, complement_notion={deduction.complement_notion!r}, "
        f"relation={deduction.relation!r})"
    )


def format_conditional_deduction_repr(deduction) -> str:
    return (
        f"ConditionalDeduction(id={deduction.id!r}, agent_id={deduction.agent_id!r}, "
        f"hypothetical_statement_id={deduction.hypothetical_statement_id!r}, "
        f"subject_notion={deduction.subject_notion!r}, complement_notion={deduction.complement_notion!r}, "
        f"relation={deduction.relation!r})"
    )



from typing import Optional, Protocol
from datetime import datetime
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.time import Time


def _to_absolute_time(time: Time) -> Optional[Time]:
    return time


class ModalContextHolder(Protocol):
    location: Optional[str]
    from_time: Optional[Time]
    to_time: Optional[Time]
    deontic_modality: Optional[object]
    quantity: Optional[float]


class ModifierMatcher:
    def matches(
        self,
        statement: ModalContextHolder,
        context: Modifier,
        reference_time: Optional[datetime] = None,
    ) -> bool:
        if context.location is not None and statement.location != context.location:
            return False
        
        if context.from_time is not None and not self._matches_time(statement.from_time, context.from_time, reference_time):
            return False
        
        if context.to_time is not None and not self._matches_time(statement.to_time, context.to_time, reference_time):
            return False
        
        if context.deontic_modality is not None:
            if statement.deontic_modality is None:
                return False
            if statement.deontic_modality != context.deontic_modality:
                return False

        if context.quantity is not None and statement.quantity != context.quantity:
            return False
        
        return True
    
    def _matches_time(self, statement_time: Optional[Time], context_time: Time, reference_time: Optional[datetime] = None) -> bool:
        if statement_time is None:
            return False
        
        resolved_statement = _to_absolute_time(statement_time)
        resolved_context = _to_absolute_time(context_time)
        
        if resolved_statement is None or resolved_context is None:
            return False
        
        if resolved_context.year is not None and resolved_statement.year != resolved_context.year:
            return False
        
        if resolved_context.month is not None and resolved_statement.month != resolved_context.month:
            return False
        
        if resolved_context.day is not None and resolved_statement.day != resolved_context.day:
            return False
        
        if resolved_context.hour is not None and resolved_statement.hour != resolved_context.hour:
            return False
        
        if resolved_context.minute is not None and resolved_statement.minute != resolved_context.minute:
            return False
        
        return True

    def exact_matches(
        self,
        statement: ModalContextHolder,
        context: Modifier,
        reference_time: Optional[datetime] = None,
    ) -> bool:
        if (context.location is None) != (statement.location is None):
            return False
        if context.location is not None and statement.location != context.location:
            return False
        if (context.from_time is None) != (statement.from_time is None):
            return False
        if context.from_time is not None and not self._matches_time(
            statement.from_time, context.from_time, reference_time
        ):
            return False
        if (context.to_time is None) != (statement.to_time is None):
            return False
        if context.to_time is not None and not self._matches_time(
            statement.to_time, context.to_time, reference_time
        ):
            return False
        if (context.deontic_modality is None) != (statement.deontic_modality is None):
            return False
        if (
            context.deontic_modality is not None
            and statement.deontic_modality != context.deontic_modality
        ):
            return False
        if (context.quantity is None) != (statement.quantity is None):
            return False
        if context.quantity is not None and statement.quantity != context.quantity:
            return False
        return True

    def context_matches(
        self,
        stored: Optional["Modifier"],
        query: Optional["Modifier"],
        reference_time: Optional[datetime] = None,
    ) -> bool:
        if query is None:
            return True
        if stored is None:
            return False
        stub = type("_Stub", (), {
            "location": stored.location,
            "from_time": stored.from_time,
            "to_time": stored.to_time,
            "deontic_modality": stored.deontic_modality,
            "quantity": stored.quantity,
        })()
        return self.matches(stub, query, reference_time)

    def context_exact_matches(
        self,
        stored: Optional["Modifier"],
        query: Optional["Modifier"],
        reference_time: Optional[datetime] = None,
    ) -> bool:
        if query is None and stored is None:
            return True
        if query is None or stored is None:
            return False
        stub = type("_Stub", (), {
            "location": stored.location,
            "from_time": stored.from_time,
            "to_time": stored.to_time,
            "deontic_modality": stored.deontic_modality,
            "quantity": stored.quantity,
        })()
        return self.exact_matches(stub, query, reference_time)



import unittest
from types import SimpleNamespace
from unittest.mock import Mock

from analogai.foundation.modules.fact.subscribers.fact_observer_registry import (
    FactObserverRegistry,
)
from analogai.foundation.modules.fact.subscribers.fact_observer_builder import (
    FactObserverBuilder,
)
from analogai.foundation.modules.fact.subscribers.statement_fact_subscriber import (
    StatementFactSubscriber,
)
from analogai.foundation.modules.fact.subscribers.notion_fact_subscriber import (
    NotionFactSubscriber,
)
from analogai.foundation.modules.fact.subscribers.hypothetical_statement_fact_subscriber import (
    HypotheticalStatementFactSubscriber,
)
from analogai.foundation.modules.fact.source.fact_source import OtherSource


class _ObservableStub:
    def __init__(self):
        self._observers = []

    def add_observer(self, observer):
        if observer not in self._observers:
            self._observers.append(observer)


class TestFactSubscribers(unittest.TestCase):
    def setUp(self):
        self.registry = FactObserverRegistry()
        self.fact_service = Mock()
        self.source = OtherSource("unit")

    def test_statement_subscriber_records_fact(self):
        subscriber = StatementFactSubscriber(self.fact_service, self.registry)
        self.registry.append_current_sentence(text="original", role="user", source=self.source)
        statement = SimpleNamespace(user_id="u1", agent_id="a1", id="s1")

        subscriber.on_statement_created(statement)

        self.fact_service.record_direct_statement_fact.assert_called_once_with(
            user_id="u1",
            agent_id="a1",
            statement_id="s1",
            text="original",
            role="user",
            user_name=None,
            timestamp=None,
            source=self.source,
        )

    def test_notion_subscriber_records_fact(self):
        subscriber = NotionFactSubscriber(self.fact_service, self.registry)
        self.registry.append_current_sentence(text="original", role="user", source=self.source)
        notion = SimpleNamespace(user_id="u1", agent_id="a1", id="n1")

        subscriber.on_notion_created(notion)

        self.fact_service.record_notion_fact.assert_called_once_with(
            user_id="u1",
            agent_id="a1",
            notion_id="n1",
            text="original",
            role="user",
            user_name=None,
            timestamp=None,
            source=self.source,
        )

    def test_hypothetical_subscriber_records_fact(self):
        subscriber = HypotheticalStatementFactSubscriber(self.fact_service, self.registry)
        self.registry.append_current_sentence(text="original", role="user", source=self.source)
        hypothetical = SimpleNamespace(user_id="u1", agent_id="a1", id="h1")

        subscriber.on_hypothetical_statement_created(hypothetical)

        self.fact_service.record_hypothetical_statement_fact.assert_called_once_with(
            user_id="u1",
            agent_id="a1",
            statement_id="h1",
            text="original",
            role="user",
            user_name=None,
            timestamp=None,
            source=self.source,
        )

    def test_builder_registers_observers_once(self):
        registry = FactObserverRegistry()
        fact_service = Mock()
        statement_service = _ObservableStub()
        notion_service = _ObservableStub()
        hypothetical_service = _ObservableStub()

        builder = FactObserverBuilder(fact_service=fact_service, registry=registry)
        builder.observe_statement_service(statement_service)
        builder.observe_statement_service(statement_service)
        builder.observe_notion_service(notion_service)
        builder.observe_hypothetical_statement_service(hypothetical_service)

        self.assertEqual(
            len([o for o in statement_service._observers if isinstance(o, StatementFactSubscriber)]),
            1,
        )
        self.assertEqual(
            len([o for o in notion_service._observers if isinstance(o, NotionFactSubscriber)]),
            1,
        )
        self.assertEqual(
            len(
                [
                    o
                    for o in hypothetical_service._observers
                    if isinstance(o, HypotheticalStatementFactSubscriber)
                ]
            ),
            1,
        )


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from typing import List, Optional

from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation


class StatementRepository(ABC):
    @abstractmethod
    def create(self, statement: DirectStatement) -> DirectStatement:
        pass

    def update(self, statement: DirectStatement) -> DirectStatement:
        return statement

    @abstractmethod
    def delete(self, statement: DirectStatement) -> None:
        pass

    @abstractmethod
    def find_by_id(self, statement_id: str) -> Optional[DirectStatement]:
        pass

    @abstractmethod
    def find_by_subject(self, subject_notion: Notion) -> List[DirectStatement]:
        pass

    @abstractmethod
    def find_by_subject_and_relation(
        self,
        subject_notion: Notion,
        relation: Relation,
        modifier: Optional[Modifier] = None,
    ) -> List[DirectStatement]:
        pass

    @abstractmethod
    def find_by_object(self, complement_notion: Notion) -> List[DirectStatement]:
        pass

    @abstractmethod
    def find_by_object_and_relation(
        self,
        complement_notion: Notion,
        relation: Relation,
        modifier: Optional[Modifier] = None,
    ) -> List[DirectStatement]:
        pass

    def find_by_agent_id(
        self,
        agent_id: str,
        limit: int = 0,
        offset: int = 0,
    ) -> List[DirectStatement]:
        return []

    def find(
        self,
        subject_notion: Notion,
        complement_notion: Notion,
        relation: Relation,
        modifier: Optional[Modifier] = None,
    ) -> Optional[DirectStatement]:
        statements = self.find_by_subject_and_relation(
            subject_notion, relation, modifier
        )
        for stmt in statements:
            if stmt.complement_notion.id == complement_notion.id:
                return stmt
        return None

    @abstractmethod
    def find_direct_statement(
        self,
        user_id: str,
        agent_id: str,
        subject_notion: Notion,
        relation: Relation,
        complement_notion: Notion,
        modifier: Optional[Modifier] = None,
    ) -> Optional[DirectStatement]:
        pass

    def find_by_captions(
        self,
        agent_id: str,
        captions: List[str],
        max_depth: int = 1,
    ) -> List[DirectStatement]:
        return []


from dataclasses import dataclass
from typing import Optional
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation


@dataclass
class BeliefExpression:
    subject_notion_expression: NotionExpression
    complement_notion_expression: NotionExpression
    relation: Relation
    probability: float
    validity: float
    modifier: Optional[Modifier] = None

    def __str__(self) -> str:
        from analogai.foundation.common.utils.expressions_serializer import serialize_belief_expression
        return serialize_belief_expression(self)

    def __repr__(self) -> str:
        return (
            f"BeliefExpression(subject_notion_expression={self.subject_notion_expression!r}, "
            f"complement_notion_expression={self.complement_notion_expression!r}, "
            f"relation={self.relation!r}, probability={self.probability!r}, "
            f"validity={self.validity!r}, modifier={self.modifier!r})"
        )



from .categorical_deduction import CategoricalDeduction
from .categorical_deduction_service import CategoricalDeductionService

__all__ = [
    "CategoricalDeduction",
    "CategoricalDeductionService",
]


from __future__ import annotations

from analogai.foundation.modules.fact.services.fact_service import FactServicePort
from analogai.foundation.modules.fact.source.repositories.fact_source_repository import FactSourceRepository
from analogai.foundation.modules.fact.source.fact_source import FactSource
from analogai.foundation.modules.fact.direct_statement_fact import DirectStatementFact
from analogai.foundation.modules.belief.belief_service import BeliefService
from analogai.foundation.common.logging.app_logger import app_logger


class FactSourceService:
    def __init__(
        self,
        fact_service: FactServicePort | None,
        belief_service: BeliefService | None,
        source_repository: FactSourceRepository | None = None,
    ) -> None:
        self._fact_service = fact_service
        self._belief_service = belief_service
        self._source_repository = source_repository

    def store_source(self, source: FactSource) -> FactSource | None:
        if source is None or self._source_repository is None:
            return None
        if not source.agent_id:
            app_logger.warning("FactSourceService store_source requires agent_id")
            return None
        self._source_repository.register_source(source)
        return source

    def get_source(self, source_id: str) -> FactSource | None:
        if not source_id or self._source_repository is None:
            return None
        return self._source_repository.get_source(source_id)

    def get_sources(self, agent_id: str) -> list[FactSource]:
        if not agent_id:
            return []
        if self._source_repository is not None:
            return self._source_repository.list_sources_for_agent(agent_id)
        return []

    def delete_source(self, source_id: str) -> dict[str, list[str]]:
        if self._fact_service is None or self._belief_service is None:
            app_logger.warning(
                "FactSourceService delete_source called without fact/belief services wired; "
                "source_id=%s",
                source_id,
            )
            raise ValueError("FactSourceService delete_source requires fact_service and belief_service")
        if not source_id:
            return {"fact_ids": [], "statement_ids": [], "belief_ids": []}
        if self._source_repository is not None:
            source = self._source_repository.get_source(source_id)
        else:
            source = None
        if source is None:
            return {"fact_ids": [], "statement_ids": [], "belief_ids": []}
        facts = self._fact_service.get_facts_by_source(source.id)
        if not facts:
            return {"fact_ids": [], "statement_ids": [], "belief_ids": []}
        turn_ids = {str(fact.turn_id) for fact in facts if fact.turn_id}
        belief_ids: list[str] = []
        for turn_id in turn_ids:
            belief_ids.extend(self._fact_service.get_belief_ids_for_turn(turn_id))
        belief_ids = [str(belief_id) for belief_id in belief_ids if belief_id]
        if source.agent_id:
            beliefs = self._belief_service.get_all_for_agent(source.agent_id)
            beliefs_by_id = {str(belief.id): belief for belief in beliefs if getattr(belief, "id", None)}
            if not belief_ids:
                statement_service = getattr(self._belief_service, "statement_service", None)
                for fact in facts:
                    if not isinstance(fact, DirectStatementFact):
                        continue
                    if statement_service is None:
                        break
                    statement = statement_service.find_by_id(fact.statement_id)
                    if statement is None:
                        continue
                    belief = self._belief_service.find(
                        statement.subject_notion,
                        statement.complement_notion,
                        statement.relation,
                    )
                    if belief is not None and getattr(belief, "id", None):
                        belief_ids.append(str(belief.id))
                belief_ids = list(dict.fromkeys(belief_ids))
            for belief_id in belief_ids:
                belief = beliefs_by_id.get(str(belief_id))
                if belief is None:
                    continue
                try:
                    self._belief_service.delete(belief)
                except Exception as exc:
                    app_logger.warning(
                        "FactSourceService failed to delete belief_id=%s source_id=%s: %s",
                        belief_id,
                        source.id,
                        exc,
                    )
        deleted_facts = self._fact_service.delete_facts_by_source(source.id)
        deleted_fact_ids = [str(fact.id) for fact in deleted_facts if fact.id]

        if self._source_repository is not None:
            try:
                self._source_repository.delete_source(source_id)
            except Exception as exc:
                app_logger.warning(
                    "Failed to delete source_id=%s from source repository: %s",
                    source_id,
                    exc,
                )

        return {
            "fact_ids": deleted_fact_ids,
            "statement_ids": [],
            "belief_ids": belief_ids,
        }



from __future__ import annotations

from typing import Optional

from analogai.foundation.settings import config as app_config
from analogai.foundation.modules.belief.belief_service import BeliefService
from analogai.foundation.modules.fact.services.fact_service import FactServicePort
from analogai.foundation.modules.fact.factories.fact_factory import FactServiceFactory
from analogai.foundation.modules.fact.repositories.infrastructure.mongo_fact_repository import (
    MongoFactRepository,
)
from analogai.foundation.setup.factories.belief_service_factory import BeliefServiceFactory
from analogai.foundation.modules.fact.source.repositories.fact_source_repository import FactSourceRepository
from analogai.foundation.modules.fact.source.repositories.infrastructure.in_memory_fact_source_repository import (
    InMemoryFactSourceRepository,
)
from analogai.foundation.modules.fact.source.repositories.infrastructure.mongo_fact_source_repository import (
    MongoFactSourceRepository,
)
from analogai.foundation.modules.fact.source.services.fact_source_service import FactSourceService


class FactSourceServiceFactory:
    def __init__(
        self,
        fact_service: Optional[FactServicePort] = None,
        source_repository: Optional[FactSourceRepository] = None,
        belief_service: Optional[BeliefService] = None,
        environment: Optional[str] = None,
    ) -> None:
        self._fact_service = fact_service
        self._source_repository = source_repository
        self._belief_service = belief_service
        self._environment = environment

    def create(self) -> FactSourceService:
        environment = self._environment or app_config.get("ENVIRONMENT", "development")
        if self._belief_service is None:
            self._belief_service = BeliefServiceFactory(environment=environment).get_service()

        statement_service = getattr(self._belief_service, "statement_service", None)
        notion_service = getattr(self._belief_service, "notion_service", None)

        if self._fact_service is None:
            self._fact_service = FactServiceFactory(
                statement_service=statement_service,
                notion_service=notion_service,
            ).create()

        if self._source_repository is None:
            repo = getattr(self._fact_service, "_repository", None)
            if isinstance(repo, MongoFactRepository):
                self._source_repository = MongoFactSourceRepository()
            else:
                self._source_repository = InMemoryFactSourceRepository()

        statement_service = getattr(self._belief_service, "statement_service", None)
        notion_service = getattr(self._belief_service, "notion_service", None)
        if statement_service is not None:
            set_stmt = getattr(self._fact_service, "set_statement_service", None)
            if callable(set_stmt):
                set_stmt(statement_service)
        if notion_service is not None:
            set_notion = getattr(self._fact_service, "set_notion_service", None)
            if callable(set_notion):
                set_notion(notion_service)

        return FactSourceService(
            fact_service=self._fact_service,
            belief_service=self._belief_service,
            source_repository=self._source_repository,
        )



from __future__ import annotations

from typing import List, Optional, Tuple
import math
import operator
import re

from analogai.foundation.settings import config as infra_config
from analogai.foundation.modules.fact.basefact import BaseFact
from analogai.foundation.modules.fact.source.fact_source import FactSource, SourceType
from analogai.foundation.modules.fact.enums.fact_reference_type import FactReferenceType
from analogai.foundation.modules.fact.repositories.fact_repository import FactRepository
from analogai.foundation.modules.fact.repositories.infrastructure.fact_storage_mapper import (
    FactStorageMapper,
)


class _RepositoryFactResolver:
    def __init__(self, repository: "FactRepository") -> None:
        self._repository = repository

    def resolve(self, fact_id: str) -> Optional[BaseFact]:
        return self._repository.get_fact(fact_id)


class MongoFactRepository(FactRepository):
    def __init__(
        self,
        uri: Optional[str] = None,
        database: Optional[str] = None,
        facts_collection: str = "facts",
        links_collection: str = "fact_links",
        turn_index_collection: str = "fact_turn_index",
    ) -> None:
        try:
            from pymongo import MongoClient
        except Exception as exc:  # pragma: no cover - optional dependency
            raise ImportError("pymongo is required for MongoFactRepository") from exc
        mongo_uri = uri or infra_config.get("MONGODB_URI", "")
        mongo_db = database or infra_config.get("MONGODB_DATABASE", "")
        if not mongo_uri or not mongo_db:
            raise ValueError("MongoDB configuration missing (MONGODB_URI/MONGODB_DATABASE)")
        client = MongoClient(mongo_uri)
        db = client[mongo_db]
        self._facts = db[facts_collection]
        self._links = db[links_collection]
        self._turn_index = db[turn_index_collection]
        self._resolver = _RepositoryFactResolver(self)
        self._ensure_indexes()

    def _ensure_indexes(self) -> None:
        try:
            self._facts.create_index([("_id", 1)], unique=True, background=True)
            self._facts.create_index([("user_id", 1), ("agent_id", 1)], background=True)
            self._facts.create_index([("turn_id", 1)], background=True)
            self._facts.create_index([("statement_id", 1)], background=True)
            self._facts.create_index([("notion_id", 1)], background=True)
            self._facts.create_index([("reference_type", 1)], background=True)

            self._links.create_index([("kind", 1), ("entity_id", 1)], unique=True, background=True)
            self._links.create_index([("turn_id", 1)], background=True)

            self._turn_index.create_index([("_id", 1)], unique=True, background=True)
            self._turn_index.create_index([("turn_ids", 1)], background=True)
        except Exception:
            pass

    def register_turn(self, user_id: str, agent_id: str, turn_id: str) -> None:
        key = f"{user_id}:{agent_id}"
        self._turn_index.update_one(
            {"_id": key},
            {"$push": {"turn_ids": turn_id}},
            upsert=True,
        )

    def save_fact(self, fact: BaseFact) -> None:
        doc = FactStorageMapper.to_document(fact)
        self._facts.update_one({"_id": fact.id}, {"$set": doc}, upsert=True)

    def get_fact(self, id: str) -> Optional[BaseFact]:
        doc = self._facts.find_one({"_id": id})
        fact = FactStorageMapper.from_document(doc)
        if fact is None:
            return None
        return self._attach_resolver(fact)

    def delete_fact(self, id: str) -> Optional[BaseFact]:
        if not id:
            return None
        doc = self._facts.find_one({"_id": id})
        if doc is None:
            return None
        self._facts.delete_one({"_id": id})
        fact = FactStorageMapper.from_document(doc)
        return self._attach_resolver(fact) if fact else None

    def update_facts_statement_id(self, turn_id: str, statement_id: str) -> None:
        if not turn_id or not statement_id:
            return
        self._facts.update_many(
            {"turn_id": turn_id},
            {"$set": {"statement_id": statement_id, "reference_type": FactReferenceType.STATEMENT.value}},
        )

    def update_facts_notion_id(self, turn_id: str, notion_id: str) -> None:
        if not turn_id or not notion_id:
            return
        self._facts.update_many(
            {"turn_id": turn_id},
            {"$set": {"notion_id": notion_id, "reference_type": FactReferenceType.NOTION.value}},
        )

    def update_facts_hypothetical_statement_id(self, turn_id: str, statement_id: str) -> None:
        if not turn_id or not statement_id:
            return
        self._facts.update_many(
            {"turn_id": turn_id},
            {"$set": {"statement_id": statement_id, "reference_type": FactReferenceType.HYPOTHETICAL_STATEMENT.value}},
        )

    def get_facts_for_turn(self, turn_id: str) -> List[BaseFact]:
        cursor = self._facts.find({"turn_id": turn_id})
        results: List[BaseFact] = []
        for doc in cursor:
            fact = FactStorageMapper.from_document(doc)
            if fact is None:
                continue
            results.append(self._attach_resolver(fact))
        return results

    def get_facts_by_source(self, source_id: str) -> List[BaseFact]:
        if not source_id:
            return []
        cursor = self._facts.find({"source_id": source_id})
        results: List[BaseFact] = []
        for doc in cursor:
            fact = FactStorageMapper.from_document(doc)
            if fact is None:
                continue
            results.append(self._attach_resolver(fact))
        return results

    def list_source_ids(self) -> List[str]:
        cursor = self._facts.find(
            {"$or": [{"source_id": {"$ne": None}}, {"source_type": {"$ne": None}}]},
            {"source_id": 1, "source_type": 1, "source_value": 1},
        )
        sources: dict[str, str] = {}
        for doc in cursor:
            source_id = _source_id_from_doc(doc)
            if not source_id:
                continue
            sources.setdefault(source_id, source_id)
        return list(sources.keys())

    def get_turn_ids(self, user_id: str, agent_id: str) -> List[str]:
        key = f"{user_id}:{agent_id}"
        doc = self._turn_index.find_one({"_id": key})
        if not doc:
            return []
        return [str(item) for item in doc.get("turn_ids", []) if item]

    def get_statement_ids_for_turn(self, turn_id: str) -> List[str]:
        if not turn_id:
            return []
        cursor = self._links.find({"kind": "statement", "turn_id": turn_id})
        return [str(doc.get("entity_id")) for doc in cursor if doc.get("entity_id")]

    def get_belief_ids_for_turn(self, turn_id: str) -> List[str]:
        if not turn_id:
            return []
        cursor = self._links.find({"kind": "belief", "turn_id": turn_id})
        return [str(doc.get("entity_id")) for doc in cursor if doc.get("entity_id")]

    def link_statement(self, statement_id: str, turn_id: str) -> None:
        if statement_id and turn_id:
            self._set_link("statement", statement_id, turn_id)

    def link_belief(self, belief_id: str, turn_id: str) -> None:
        if belief_id and turn_id:
            self._set_link("belief", belief_id, turn_id)

    def link_notion(self, notion_id: str, turn_id: str) -> None:
        if notion_id and turn_id:
            self._set_link("notion", notion_id, turn_id)

    def link_hypothetical_statement(self, statement_id: str, turn_id: str) -> None:
        if statement_id and turn_id:
            self._set_link("hypothetical_statement", statement_id, turn_id)

    def get_turn_id_for_statement(self, statement_id: str) -> Optional[str]:
        return self._get_link("statement", statement_id)

    def get_turn_id_for_belief(self, belief_id: str) -> Optional[str]:
        return self._get_link("belief", belief_id)

    def get_turn_id_for_notion(self, notion_id: str) -> Optional[str]:
        return self._get_link("notion", notion_id)

    def get_turn_id_for_hypothetical_statement(self, statement_id: str) -> Optional[str]:
        return self._get_link("hypothetical_statement", statement_id)

    def search_text(self, user_id: str, agent_id: str, query: str, top_k: int) -> List[BaseFact]:
        safe = re.escape(query)
        cursor = self._facts.find(
            {
                "user_id": user_id,
                "agent_id": agent_id,
                "text": {"$regex": safe, "$options": "i"},
            }
        ).limit(max(1, int(top_k)))
        results: List[BaseFact] = []
        for doc in cursor:
            fact = FactStorageMapper.from_document(doc)
            if fact is None:
                continue
            results.append(self._attach_resolver(fact))
        return results

    def search_similar(
        self,
        user_id: str,
        agent_id: str,
        embedding: List[float],
        top_k: int,
        similarity_cutoff: Optional[float] = None,
    ) -> List[BaseFact]:
        cursor = self._facts.find(
            {
                "user_id": user_id,
                "agent_id": agent_id,
                "embedding": {"$ne": None},
            }
        )
        scored: List[Tuple[float, BaseFact]] = []
        for doc in cursor:
            fact = FactStorageMapper.from_document(doc)
            if fact is None:
                continue
            fact = self._attach_resolver(fact)
            if not fact.embedding:
                raise RuntimeError("Fact missing embedding")
            score = _cosine_similarity(embedding, fact.embedding)
            if similarity_cutoff is not None and score < similarity_cutoff:
                continue
            scored.append((score, fact))
        scored.sort(key=operator.itemgetter(0), reverse=True)
        return [fact for _, fact in scored[: max(1, int(top_k))]]

    def delete_facts_by_source(self, source_id: str) -> List[BaseFact]:
        if not source_id:
            return []
        query = {"source_id": source_id}
        cursor = self._facts.find(query)
        deleted: List[BaseFact] = []
        for doc in cursor:
            fact = FactStorageMapper.from_document(doc)
            if fact is None:
                continue
            deleted.append(self._attach_resolver(fact))
        if deleted:
            self._facts.delete_many(query)
        return deleted

    def _set_link(self, kind: str, entity_id: str, turn_id: str) -> None:
        key = f"{kind}:{entity_id}"
        self._links.update_one(
            {"_id": key},
            {"$set": {"turn_id": turn_id, "kind": kind, "entity_id": entity_id}},
            upsert=True,
        )

    def _get_link(self, kind: str, entity_id: str) -> Optional[str]:
        key = f"{kind}:{entity_id}"
        doc = self._links.find_one({"_id": key})
        if not doc:
            return None
        return doc.get("turn_id")

    def _attach_resolver(self, fact: BaseFact) -> BaseFact:
        fact._resolver = self._resolver
        fact._deleter = self.delete_fact
        return fact


def _cosine_similarity(a: List[float], b: List[float]) -> float:
    if not a or not b:
        return 0.0
    if len(a) != len(b):
        return 0.0
    dot = sum(x * y for x, y in zip(a, b))
    norm_a = math.sqrt(sum(x * x for x in a))
    norm_b = math.sqrt(sum(y * y for y in b))
    if norm_a == 0.0 or norm_b == 0.0:
        return 0.0
    return dot / (norm_a * norm_b)


def _source_id_from_doc(doc: dict) -> Optional[str]:
    if not doc:
        return None
    source_id = doc.get("source_id")
    if source_id:
        return str(source_id)
    agent_id = doc.get("agent_id")
    source_type = doc.get("source_type")
    source_value = doc.get("source_value")
    if source_type is None or source_value is None:
        return None
    try:
        stype = SourceType(str(source_type))
    except Exception:
        try:
            stype = SourceType[str(source_type).upper()]
        except Exception:
            stype = SourceType.OTHER
    source = FactSource(stype, str(source_value), str(agent_id) if agent_id else None)
    return source.id
